// material 3d interface

/* material flags
 1 - n
 2 - uv
 4 - colors
 8 - tm
 16 - bump
 32 - uv2
 64 - color emissive
 256 - selected
 512 - double sided
 1024 - just transparency
 2048 - shadows
- 0x30000 - z offset
*/

iv.mtlchannel=function(){this.mode=0;this.color=null;this.texture=null;}
iv.mtlvar=function(id,ord){this.id=id;this.ord=ord;this.slot=null;}

iv.shader=function(m,f) {
	this.mtl=m;
	this.flags=f;
	this._reset();
};
iv.shader.prototype._reset=function()
{
	this.valid=false;
	this.attrs=[];// list of attrs to program
	this.vars=[];// list of variable to program
	this.textures=[];
	this.program=null;
	this.vShader=null;
	this.fShader=null;
	this.loadedtextures=0;
	this.numLights=0;
}

iv.material=function(space){
this.space=space;
this.spaceId=space.spaceId;
this.gl=space.gl;
this.shaders=[];
this.phong=18;
}

iv.mtlchannel.prototype.setColor=function(clr) {
	if(!this.color) this.color=[0,0,0];
	var c=this.color,t=typeof clr;
	if(t==='number') {
		c[0]=((clr>>16)&0xff)/255;
		c[1]=((clr>>8)&0xff)/255;
		c[2]=(clr&0xff)/255;
	}
	else{c[0]=clr[0];c[1]=clr[1];c[2]=clr[2];}
}

iv.mtlchannel.prototype.setTime=function(t)
{
	if(this.anim)
	{
		var a=this.anim;
		if(a.amount)this.amount=iv.getFloatTrackInfo(a.amount,t,this.amount);
		if(a.color){var c=iv.getP3DTrackInfo(a.color,t);if(c)this.setColor(c);}
	}
}

iv.mtlchannel.prototype.activateAnimation=function(a,reset)
{
	if(this.anims)
		iv.anim.activate(this,a,reset);
}
iv.mtlchannel.prototype.replaceAnimId=function(a,b)
{
	if(this.anims)
		iv.anim.replace(this,a,b);
}

iv.material.prototype.setTime = function(t)
{
	for(var i=0;i<iv.mtlChannels.length;i++)
	{
		var c=this[iv.mtlChannels[i]];
		if(c)
		{
			for(var j=0;j<c.length;j++)c[j].setTime(t);
		}
	}
}
iv.material.prototype.replaceAnimId=function(a,b)
{
	for(var i=0;i<iv.mtlChannels.length;i++)
	{
		var c=this[iv.mtlChannels[i]];
		if(c)
		{
			for(var j=0;j<c.length;j++)c[j].activateAnimation(a,reset);
		}
	}
}
iv.material.prototype.activateAnimation = function(a,reset)
{
	for(var i=0;i<iv.mtlChannels.length;i++)
	{
		var c=this[iv.mtlChannels[i]];
		if(c)
		{
			for(var j=0;j<c.length;j++)c[j].activateAnimation(a,reset);
		}
	}
}

iv.material.prototype.invalidate=function() {
	var s=this.shaders,i;
	if(s){
		for(i=0;i<s.length;i++)
			s[i].detach(this.gl);
		this.shaders=[];
	}
}

iv.material.prototype.clear = function()
{
	for(var i=0;i<iv.mtlChannels.length;i++)
	{
		var c=iv.mtlChannels[i];
		if(this[c])this[c]=null;
	}
}

iv.material.prototype.isChannel  = function(c)
{
	if( (c===undefined)||(c===null))return false;
	if(c.length===0)return false;
	for(var i=0;i<c.length;i++)
	{
		var item=c[i];
		if(item.texture!=null || item.color!=null || item.amount!=null)return true;
	}
	return false;
}

iv.material.prototype.newChannel = function(type,ch)
{
	if(!ch)ch=new iv.mtlchannel();
	if(!(type in this))this[type]=[];
	this[type].push(ch);
	this.valid=false;
	return ch;
}

iv.material.prototype.getChannel = function(type)//returns first channel
{
	if(!(type in this))return null;
	var items=this[type];
	return items[0];
}


iv.material.prototype.newTexture = function(c,name,type)
{
	var gl=this.gl;
	if(type===undefined)type=gl.TEXTURE_2D;
	c.texture=this.space.getTexture(name,type,this);
	if(type==gl.TEXTURE_CUBE_MAP)c.wrapT=c.wrapS=gl.CLAMP_TO_EDGE;
	else
	{
		if(!c.wrapS)c.wrapS=gl.REPEAT;
		if(!c.wrapT)c.wrapT=gl.REPEAT;
	}
}

iv.material.prototype.cnvTtxMatrix=function (a)
{
	var tm=mat3.create(),index=0,i,j;
	for(i=0;i<3;i++)
	{
		for(j=0;j<2;j++){tm[i*3+j]= a[index];index++;}
	}
	tm[2]=0;tm[5]=0;tm[8]=1;
	return tm;
}

iv.material.prototype.loadChannelImp=function(v,name) {
	var c=this.newChannel(name),t=v.texture;
	if(v.color!==undefined) c.setColor(v.color);
	if(v.amount!==undefined) c.amount=v.amount;
	if(v.blend) c.blend=v.blend;
	if(!t && v.inline)t="?s"+this.spaceId+"i"+v.inline;
	if(t!==undefined){
		var type;
		if(v.type&&v.type=="cube") {
			if(this.gl) type=this.gl.TEXTURE_CUBE_MAP;
			else type=0x8513;// double check this
		}
		if(v.tm)
			c.tm=this.cnvTtxMatrix(v.tm);
		if(v.cmp) c.cmp=v.cmp;
		if(v.filter) c.filter=v.filter;
		if(v.uvset)c.uv=v.uvset;else c.uv=0;
		if(v.wrapT) c.wrapT=v.wrapT;
		if(v.wrapS) c.wrapS=v.wrapS;
		if(v.format) c.format=v.format;
		this.newTexture(c,t,type);
	}

	if(v.anims)c.anims=v.anims;

}

iv.material.prototype.loadChannel=function(v,name) {
	var type=typeof v;
	if(type==="number") {
		var c=this.newChannel(name);
		if(name=='opacity') c.amount=v; else c.setColor(v);
	}else
		if(type==="object") {
			if(v instanceof Array) {
				var len=v.length;
				if((len==3)&&(typeof v[0]=='number')&&(typeof v[1]=='number')&&(typeof v[2]=='number')) {
					var c=this.newChannel(name);
					c.setColor(v);
				} else {
					for(var i=0;i<len;i++) this.loadChannelImp(v[i],name);
				}
			} else this.loadChannelImp(v,name);
		}
};

iv.material.prototype.load=function(d)
{
	for(var v in d)
	{
	var a=d[v];
	switch(v)
	{
	case "lightmap":
	case "diffuse":
	case "specular":
	case "emissive":
	case "reflection":
	case "opacity":
	case "bump":this.loadChannel(a,v);break;
	case "ambient":this.loadChannel(a,"emissive");break;
	case "name":
	case "phong":
	case "backSide":
	case "path":this[v]=a;break;
	}}
	return true;
}


iv.material.prototype.getShader = function(flags)
{
	if(!this.staticColors)flags&=~256;// remove selection
	if(!this.space.cfgTextures)flags&=~2;
	// we may need to remove bits from flags
	for(var i=0;i<this.shaders.length;i++)
	{
		var s=this.shaders[i];
		if(s.flags==flags)
		{
			if((s.loadedtextures!=s.textures.length) && s.valid)
			{
				var c=s.readyTextures(false);
				if(c!=s.loadedtextures)s.valid=false;
			}
			if(s.numLights!=this.space.lights.length)s.valid=false;
			return s;
		}
	}
	var s=new iv.shader(this,flags);
	this.shaders.push(s);
	return s;
}


iv.shader.prototype.addVar = function(id,ord)
{
	var v=new iv.mtlvar(id,ord);
	this.vars.push(v);
	return v;
}

iv.shader.prototype.addAttr = function(id,shName,gl)
{
	var attr={};
	attr.id=id;
	attr.slot= gl.getAttribLocation(this.program, shName);
	gl.enableVertexAttribArray(attr.slot);// do we need this
	this.attrs.push(attr);
}

iv.shader.prototype.addLightVar = function(id,name,light,ord,t,ft)
{
	var v=this.addVar(id,ord);
	v.name=name;
	v.light=light;
	ft.push("uniform "+t+" "+name+";");
	return v;
}

iv.shader.prototype.addChVar = function(id,name,ch,ord)
{
	var v=this.addVar(id,ord);
	v.name=name;
	v.channel=ch;
	return v;
}

iv.shader.prototype.compareTM3=function (a,b)
{
	if(a===undefined && b===undefined)return true;
	if(a===undefined || b===undefined)return false;
	for(var i=0;i<9;i++)
		if(Math.abs(a[i]-b[i])>1e-4)return false;
	return true;
}

iv.shader.prototype.getTexture=function(c)
{
	var items=this.textures,j,t;
	for(j=0;j<items.length;j++)
	{
		t=items[j];
		if(t.txt===c.texture && (t.wrapS==c.wrapS) && (t.wrapT==c.wrapT) && this.compareTM3(t.tm,c.tm) && c.uv==t.uv)
			return t;
	}
	return null;
}
iv.shader.prototype.num = function(a)
{
	var t=a.toString();
	if(t.indexOf('.')<0)t+='.0';
	return t;
}
iv.shader.prototype.num3 = function(a)
{
return "vec3("+this.num(a[0])+","+this.num(a[1])+","+this.num(a[2])+")";
}

iv.shader.prototype.preChannel=function(ch,ft)
{
	var text="",i,c;
	for(i=0;i<ch.length;i++)
	{
		c=ch[i];
		c._id=this.channelId;this.channelId++;
		if(c.color!=null)
		{
			var name="ch"+c._id+"clr";
			if(this.mtl.staticColors)
			{
				var clr=c.color;
				if(this.flags&256)
				{
					var _s=this.mtl.space.clrSelection;
					clr=vec3.lerp(c.color,_s,_s[3]);
				}
				text+="const vec3 "+name+"="+this.num3(clr)+";" ;
			}else{
				this.addChVar("color",name,c,4102);
				text+="uniform vec3 "+name+";";
			}
		}
		if("amount" in c)
		{
			var name="ch"+c._id+"amount";
			if(this.mtl.staticColors)
			{
				text+="const float  "+name+"="+this.num(c.amount)+";";
			}else{
				this.addChVar("amount",name,c,4104);
				text+="uniform float "+name+";";
			}
		}
	}
	ft.push(text);
};


iv.shader.prototype.handleChannel_=function(gl,ch,ft)
{
	var text="",text2="",i,c;
	for(i=0;i<ch.length;i++)
	{
		var c=ch[i],cname=null,tname=null,aname=null;
		if(c.color!=null){cname="ch"+c._id+"clr";}
		if(c.texture!=null && c.texture.ready)
		{
			var t=this.getTexture(c);
			if(t){
				if(c.texture.isCube())
				{
					text+="vec3 lup = reflect(eyeDirection,normal);lup.y*=-1.0;vec4 refColor="+"textureCube(txtUnit"+t.slot+",lup);";//sorry - only one reflection texture, remove lup.y-1 from here
					tname="refColor";
				}else{
					tname="txtColor"+t.slot;
				}
			}
		}
		if(tname && c.amount!=null){aname="ch"+c._id+"amount";}
		if(cname || tname)
		{
			var local=null;
			if(aname && tname){local="vec3("+aname+")*vec3("+tname +")";
				if(cname)local+="*"+cname;
			}else
				if(cname && tname)local=cname+"*vec3("+tname +")";// color by texture
				else
					if(cname)local=cname;
					else local="vec3("+tname+")";
			if(i && c.texture)
			{
				if(c.format=='rgba'){
					if(aname)aname=aname+'*'+tname+'.a';
					else
						aname=tname+'.a';
				}}
			if(text2.length)
			{
				if(c.blend=='blend' && aname)text2="mix("+text2+","+local+','+aname+')';
				else
					text2="("+text2+")"+this.getBlend(c)+local;
			}else text2=local;
		}
	}
	if(text.length)ft.push(text);
	return text2;
}

iv.shader.prototype.handleChannel=function(gl,ch,cmp,ft)
{
	var text2=this.handleChannel_(gl,ch,ft);
	if(text2){
		if(cmp)text2=cmp+"*="+text2;
		else text2="color+="+text2;
		ft.push(text2+";");
	}
	if(cmp)ft.push("color+="+cmp+";");
};

iv.shader.prototype.getBlend=function(c) {
	switch(c.blend) {
		case "sub":return "-";
		case "mul":return "*";
		default:return "+";
	}
}

iv.shader.prototype.handleAlphaChannel=function(gl,ch,ft)
{
	if(ch && ch.length)
	{
		var txt=null,tAlpha=false;
		for(var i=0;i<ch.length;i++)
		{
			var c=ch[i],tname=null,aname=null;
			if(c.texture && c.texture.ready)
			{
				var t=this.getTexture(c);
				if(t)
					tname="txtColor"+t.slot;
			}
			if(c.amount!=null){aname="ch"+c._id+"amount";}
			if(tname)
			{	
				if(aname)t=aname+"*";else t="";
				if(c.cmp && c.cmp=='a')
					t+=tname+".a";
				else
					t+="("+tname+".x+"+tname+".y+"+tname+".z)/3.0";
			}else
				if(aname)t= aname;
			if(t)
			{
				if(txt)
				{
					if(!tAlpha)
					{
						txt="float alpha="+txt+";\n";
						tAlpha=true;
					}
					txt+="alpha"+ this.getBlend(c)+'='+t+";\n";
				}else txt=t;
			}
		}
		if(txt)
		{
			if(tAlpha)
			{
				ft.push(txt);
				return "alpha";
			}else return txt;
		}
	}
	return "1.0";
}

iv.shader.prototype.handleBumpChannel=function(gl,ch)
{
	if(ch && ch.length)
	{
		var c=ch[0];
		if(c.texture && c.texture.ready)
		{
			var t=this.getTexture(c);
			if(t){
					var tname="txtColor"+t.slot,text="vec3 _n=vec3("+tname+");";
					text+="_n-=vec3(0.5,0.5,0);_n*=vec3(2.0,2.0,1.0);";
					if(c.amount!=null){var aname="ch"+c._id+"amount";text+="_n*=vec3("+aname+","+aname+",1.0);";}
					text+="_n=normalize(_n);";
					return text;
			}
		}
	}
	return null;
}

iv.shader.prototype.collectTextures=function(ch,f)
{
	var rez=0,i,c;
	if(ch)
	{
		for(i=0;i<ch.length;i++)
		{
			c=ch[i];
			if(c.texture&& ( (f<0)|| (f&(1<<c.uv)  ) ) ){
				if(!this.getTexture(c))
				{
					var t={txt:c.texture,slot:0,wrapS:c.wrapS,wrapT:c.wrapT,uv:c.uv?c.uv:0};
					if(c.tm){t.tm=c.tm;
					t.ch=c;
					}
					this.textures.push(t);
				}
				if(c.texture.ready)
					rez|=1<<c.uv;
			}
		}
	}
	return rez;
}

iv.shader.prototype.readyTextures = function(set)
{
	var c=0,i,t;
	for(i=0;i<this.textures.length;i++)
	{
		t=this.textures[i];
		if(t.txt.ready)
		{
			if(set)t.slot=c;
			c++;
		}
	}
	return c;
}

iv.shader.prototype.compile=function(i,type)
{
	return this.mtl.space.compileShader(i.join(''),type);
}

iv.shader.prototype.fetchTextures=function(gl,ft)
{
	var _i=this.textures,a=0,i,t,s;
	for(i=0;i<_i.length;i++)
	{
		t=_i[i];
		if(t.txt.ready && !t.light && t.txt.is2d()){
			if(t.tm){
				s="_uv=vec2(ch"+t.ch._id+"tm*vec3(vUV"+t.uv+",1.0));\n";
				ft.push((a?"":"vec2 ")+s);
				a++;//counter of textures
			}
			ft.push("vec4 txtColor"+t.slot+"= texture2D(txtUnit"+t.slot+","+ ((t.tm)?"_uv":"vUV"+t.uv)+");");
		}
	}
}

iv.shader.prototype.update = function()
{	
	var mtl=this.mtl,gl=mtl.gl,i;
	if(this.program)this.detach(mtl.gl);// may be just update?
	this.numLights=mtl.space.lights.length;
	this.channelId=0;
	
	var _lights=null,vt=[],ft=[];
	if(this.flags&8)vt.push("uniform mat4 tmWorld,tmModelView,tmPrj;");
	if(this.flags&iv.R_Z_OFFSET)
		vt.push("uniform float zOffset;");
	var bNormals=(this.flags&1)!=0,bSpecular=false,bDiffuse=false,bLights=false,bReflection=false,bBump=false,bLightMap=false;
	var bEmissive=mtl.isChannel(mtl.emissive);

	var bOpacity=mtl.isChannel(mtl.opacity) || (this.flags&1024);
	var lights=mtl.space.lights;
	vt.push("attribute vec3 inV;");
	if(this.flags&8)vt.push("varying vec4 wPosition;");

	if(this.flags&4)//diffuse colors
		vt.push("varying vec3 vC;attribute vec3 inC;");

	if(this.flags&64)// emissive colors
		vt.push("varying vec3 vCE;attribute vec3 inCE;");
	var bUV=0;

	if(bNormals){
		vt.push("varying vec3 wNormal;attribute vec3 inN;");
		if(lights.length)
		{
			if(mtl.isChannel(mtl.diffuse))bDiffuse=true;
			if(mtl.isChannel(mtl.specular))bSpecular=true;
			bLights=bDiffuse||bSpecular;
		}
	if(mtl.space.cfgTextures)
		bReflection=this.collectTextures(mtl.reflection,-1);
	}
	var _uvflags=0;
	if(this.flags&2)_uvflags|=1;
	if(this.flags&32)_uvflags|=2;
	if(_uvflags){
		if(bDiffuse)bUV|=this.collectTextures(mtl.diffuse,_uvflags);
		if(bSpecular)bUV|=this.collectTextures(mtl.specular,_uvflags);
		if(bEmissive)bUV|=this.collectTextures(mtl.emissive,_uvflags);
		if(bOpacity)bUV|=this.collectTextures(mtl.opacity,_uvflags);
		if(bNormals && bLights)bUV|=(bBump=this.collectTextures(mtl.bump,_uvflags));
		bUV|=(bLightMap=this.collectTextures(mtl.lightmap,_uvflags));
	}
	if(bUV&1)vt.push("varying vec2 vUV0;attribute vec2 inUV0;");
	if(bUV&2)vt.push("varying vec2 vUV1;attribute vec2 inUV1;");

	if(bBump)
	{vt.push("varying vec3 vBN,vBT;attribute vec3 inBN,inBT;");}
		
	vt.push("void main(void){");
	
	if(this.flags&8){
		vt.push("wPosition= tmWorld*vec4(inV,1.0);vec4 vPosition=tmModelView*wPosition;gl_Position=tmPrj*vPosition;");
		this.addVar("tmWorld",4101);this.addVar("tmModelView",4114);this.addVar("tmPrj",4115);
	}
	else vt.push("gl_Position = vec4(inV,1.0);");

	if(this.flags&iv.R_Z_OFFSET)// z offset
	{
		this.addVar("zOffset",4105);
		vt.push("gl_Position.z+=zOffset;");
	}
	
	if(bNormals){
		vt.push("wNormal = ");
		if(mtl.backSide)vt.push("-");
		vt.push("normalize(vec3(tmWorld* vec4(inN,0.0)));");
	}
	if(bBump)vt.push("vBN=normalize(vec3(tmWorld*vec4(inBN,0.0)));vBT=normalize(vec3(tmWorld*vec4(inBT,0.0)));");
	if(bUV&1)vt.push("vUV0=inUV0;");
    if(bUV&2)vt.push("vUV1=inUV1;");
	if(this.flags&4)vt.push("vC = inC;");
	if(this.flags&64)vt.push("vCE = inCE;");
	vt.push("}");
	
	ft.push("precision mediump float;");
	if(this.flags&1024)ft.push("uniform float opacity;");
	if(bNormals )ft.push("varying vec4 wPosition;");
	if(this.flags&4)ft.push("varying vec3 vC;");
	if(this.flags&64)ft.push("varying vec3 vCE;");
	if(bUV&1)ft.push("varying vec2 vUV0;");
	if(bUV&2)ft.push("varying vec2 vUV1;");

	if(bBump)ft.push("varying vec3 vBN,vBT;");
	if(bDiffuse)this.preChannel(mtl.diffuse,ft);
	if(bSpecular)this.preChannel(mtl.specular,ft);
	if(bEmissive)this.preChannel(mtl.emissive,ft);
	if(bReflection)this.preChannel(mtl.reflection,ft);
	if(bOpacity&& mtl.opacity)this.preChannel(mtl.opacity,ft);
	if(bBump)this.preChannel(mtl.bump,ft);



	if(bNormals){
		ft.push("uniform vec3 eye;");
		this.addVar("eye",4113);
		if(bSpecular)
		{
			ft.push("uniform float mtlPhong;");
			this.addVar("mtlPhong",4116);
		}
		ft.push("varying vec3 wNormal;");
		ft.push("float k;");
		if(bLights)
		{
		ft.push("vec3 diffuse,specular,lightDir;");//lightDir - for point lights only
		_lights=[];
		var anySpot=false,shI=0,shS=0,shP=0;
		for(i=0;i<lights.length;i++)
		{
			var ls=lights[i],colorname="light"+i+"Clr",l={light:ls,type:ls.type,colorname:colorname};
			this.addLightVar("lightColor",colorname,ls,4110,"vec3",ft);

			if(ls.dir)
				this.addLightVar("lightDir",l.dirname="light"+i+"Dir",ls,4112,"vec3",ft);
			if(ls.spot)
			{
				anySpot=true;
				l.spotname="light"+i+"Spot";
				this.addLightVar("lightSpot",l.spotname,ls,4117,"vec3",ft);
			}
			if(ls.org)
			{
				if(ls.org==='camera')
				{
					l.orgname="eye";
					l.type =0;
				}else	this.addLightVar("lightOrg",l.orgname="light"+i+"Org",ls,4111,"vec3",ft);
			}

			if(ls.shadowOK && (this.flags&2048))
			{
				l.biasname="light"+i+"bias";
				this.addLightVar("lightProps",l.biasname,ls,4132,"vec4",ft);
				var shX=0;
				switch(ls.light.shadowFilter)
				{
				case 5:shX=4;break;
				case 3:shX=2;break;
				default:shX=1;break;
				}
				if(ls.type==0){shP|=shX;l.vtmname="*";}else
				{
					if(ls.type==2)shS|=shX;else shI|=shX;
					l.vtmname="light"+i+"vtm";
					l.ptmname="light"+i+"ptm";
					this.addLightVar ("lightVTM",l.vtmname,ls,4130,"mat4",ft);
					this.addLightVar ("lightPTM",l.ptmname,ls,4131,"mat4",ft);
					if(shX>1)
					{
						this.addLightVar ("lightV",l.uvVname="light"+i+"V",ls,4133,"vec3",ft);
						this.addLightVar ("lightU",l.uvUname="light"+i+"U",ls,4134,"vec3",ft);
					}
				}
				var t={txt:ls.lightDepthTexture,slot:0,wrapS:gl.CLAMP_TO_EDGE,wrapT:gl.CLAMP_TO_EDGE,uv:-1,light:l};
				this.textures.push(t);
				l.texture=t;
			}else
			{
				if(l.vtmname)l.vtmname=null;
			}
 
			_lights.push(l);
		}
		if(anySpot||shI||shP||shS)ft.push("float angle,bias;");
		}
	}
	// after lights
	this.loadedtextures=this.readyTextures(true);
	for(i=0;i<this.textures.length;i++)
	{
		var t=this.textures[i];
		if(t.txt.ready){
			ft.push("uniform "+(t.txt.isCube()?"samplerCube":"sampler2D")+" txtUnit"+ t.slot+";");
			if(t.tm){
				var v=this.addVar("tm",4103);
				v.channel=t.ch;
				ft.push("uniform mat3 ch"+t.ch._id+"tm;");
			}
		}
	}

if(shI||shS)
{
	ft.push("float texture2DCompare(sampler2D depths, vec2 uv, float compare){float depth = texture2D(depths, uv).r;return step(depth,compare);}");
	if(shI&4)this.genISampler(ft,5);
	if(shI&2)this.genISampler(ft,3);
	if(shS&4)this.genSSampler(ft,5);
	if(shS&2)this.genSSampler(ft,3);
	if(shI&1)this.gen2dSampler1(ft,false);
	if(shS&1)this.gen2dSampler1(ft,true);
}
	if(shP&4)this.gen3dSampler(ft,5);
	if(shP&2)this.gen3dSampler(ft,3);
	if(shP&1)this.gen3dSampler1(ft);


	ft.push("void main(void){");
    this.fetchTextures(gl,ft);

	if(bNormals)
	{
		ft.push("vec3 normal = normalize(wNormal);");
		var normalSign="";
		if(this.flags&512){ft.push("vec3 normalSign=vec3(gl_FrontFacing?1.0:-1.0);");normalSign="normalSign*";}

		if(bBump)
		{
			var txt=this.handleBumpChannel(gl,mtl.bump);
			if(txt){
				ft.push(txt);
				ft.push("mat3 tsM = mat3(normalize("+normalSign+"vBN), normalize("+normalSign+"vBT), normal);");
				ft.push("normal =  normalize(tsM*_n);");
			}
		}
		if(this.flags&512)ft.push("normal="+normalSign+"normal;");
		
		ft.push("vec3 eyeDirection = normalize(wPosition.xyz-eye);vec3 reflDir;float specA;");
if(shP||shP||shI)ft.push("vec3 posInLight;float lightDepth,depth;");
		if(_lights)
		for(i=0;i<_lights.length;i++)
		{
			var l=_lights[i];
			
			var imageW=1;if(l.vtmname)imageW=l.light.lightDepthTexture.width;
			if(l.light.shaders)
			{
				if(iv.indexOf(l.light.shaders,this)<0)l.light.shaders.push(this);
			}else l.light.shaders=[this];
			
			var dirName,kstr="k*";
			switch(l.type)
			{
//Point
			case 0:dirName="lightDir";

			if(l.vtmname)
			{   
				kstr="k*angle*";
				ft.push("posInLight = wPosition.xyz-"+l.orgname+";");
				ft.push("lightDir=normalize(posInLight);");
				this.f0(ft,l,"length(posInLight)");
				ft.push("angle="+this.fetchShadow3d(l,imageW));
			}else ft.push("lightDir=normalize(wPosition.xyz-"+l.orgname+");");
			break;
//Infinite
			case 1:dirName=l.dirname;

				if(l.vtmname)
				{
					kstr="k*angle*";
					ft.push("\nangle="+this.fetchShadowI(l)+"\n");
				}break;
//Spot
			case 2:
				ft.push("lightDir=normalize( wPosition.xyz-"+l.orgname+");");
				ft.push("angle=smoothstep("+l.spotname+"[1],"+l.spotname+"[0],dot(lightDir,"+l.dirname+"));");
				kstr="k*angle*";
				dirName="lightDir";

			if(l.vtmname)
			{
				ft.push("\nangle*="+this.fetchShadowS(l)+"\n");
			} break;
			}
			if(bSpecular)
			{
			ft.push("reflDir=reflect(-"+dirName+", normal);specA=dot(reflDir,eyeDirection);k=pow(max(specA,0.0),mtlPhong);");
			if(i)ft.push("specular+=");else ft.push("specular=");
			ft.push(kstr+l.colorname+";");
			}
			if(bDiffuse)
			{
			ft.push("k=max(dot(normal, -"+dirName+"), 0.0);");
			if(i)ft.push("diffuse+=");else ft.push("diffuse=");
			ft.push(kstr+l.colorname+";");
			}
		}
	}
	ft.push("vec3 color= vec3(0.0);");

	if(this.flags&4)
	{
		if(bDiffuse && (!(this.flags&64)))ft.push("diffuse=diffuse*vC;");
		else
		{
			ft.push("color+=vC;");
			bEmissive=false;
		}
	}
	if(this.flags&64)ft.push("color+=vCE;");
	if(bDiffuse)
		this.handleChannel(gl,mtl.diffuse,"diffuse",ft);
	
	if(bSpecular)
		this.handleChannel(gl,mtl.specular,"specular",ft);
	if(bLightMap)
	{
		var aoText=this.handleChannel_(gl,mtl.lightmap,ft);
		if(aoText)ft.push("color*="+aoText+";");
	}

	if(bEmissive)
	this.handleChannel(gl,mtl.emissive,null,ft);
	var alpha="1.0";

	if(bOpacity)
	{
		var n;
		if(this.flags&1024)this.addVar("opacity",4106);
		if(mtl.opacity){
			n=this.handleAlphaChannel(gl,mtl.opacity,ft);
			if(this.flags&1024){ft.push("float _opacity=opacity*"+n+";");n="_opacity";}
		}else n="opacity";
		if(bReflection)
		{
		var text2=this.handleChannel_(gl,mtl.reflection,ft);
		ft.push("vec3 _refColor="+text2+";");
		ft.push("float _refA=(_refColor.x+_refColor.y+_refColor.z)/3.0;");
		ft.push("_refA=1.0-(1.0-_refA)*(1.0-"+n+");");
		n="_refA";
		ft.push("color+=_refColor;");
		}else ft.push("if("+n+"<0.004)discard;");
		ft.push("gl_FragColor = vec4(color,"+n+");");
		
	}else {
	if(bReflection)this.handleChannel(gl,mtl.reflection,null,ft);
		ft.push("gl_FragColor=vec4(color,1.0);");
	}
	ft.push("}");
	this.vShader = this.compile(vt,gl.VERTEX_SHADER);
	this.fShader = this.compile(ft,gl.FRAGMENT_SHADER);
	var shPrg=this.program=gl.createProgram();
	gl.attachShader(shPrg, this.vShader);
	gl.attachShader(shPrg, this.fShader);
	gl.linkProgram(shPrg);

	if (!gl.getProgramParameter(shPrg, gl.LINK_STATUS)) {
		alert("Could not initialise shaders");
	}

	gl.useProgram(shPrg);

	// addAttr after useProgram
	this.addAttr(4300,"inV",gl);
	if(bNormals)this.addAttr(4301,"inN",gl);
	if(bUV&1)this.addAttr(4302,"inUV0",gl);
	if(bUV&2)this.addAttr(4306,"inUV1",gl);
	if(bBump){this.addAttr(4303,"inBN",gl);this.addAttr(4304,"inBT",gl);}
	

	if(this.flags&4)this.addAttr(4305,"inC",gl);
	if(this.flags&64)this.addAttr(4307,"inCE",gl);

	for(i=0;i<this.textures.length;i++)
	{
		var t=this.textures[i];
		if(t.txt.ready)
			t.uniform=gl.getUniformLocation(shPrg,"txtUnit"+t.slot);
	}
	i=0;
	while(i<this.vars.length)
	{
		var v=this.vars[i],name=null;
		switch(v.id)
		{
		case "tm":name= "ch"+v.channel._id+"tm";break;
		case "color":name= "ch"+v.channel._id+"clr";break;
		case "amount":name= "ch"+v.channel._id+"amount";break;

		case "lightVTM":
		case "lightPTM":
		case "lightProps":
		case "lightV":
		case "lightU":

		case "lightSpot":
		case "lightColor":
		case "lightDir":
		case "lightOrg":name=v.name;break;
		default:name=v.id;
		}
		v.slot=gl.getUniformLocation(shPrg, name);
		if(!v.slot)
		{
			this.vars.splice(i,1);
		// console.log("Slot '"+name+"' not found"); // uncomment for debug reasons
		}else i++;
	}
	this.valid=true;
	return true;
}


iv.shader.prototype.fetchShadow3d=function(l,imageW)
{
	var filter=l.light.light.shadowFilter,str;
	if(filter>1)
		str="Shadow3D"+filter+"(txtUnit"+l.texture.slot+","+(1.0/imageW)+",lightDir,wPosition.xyz,"+l.orgname+",normal,depth,"+l.biasname+")";
	else str="Shadow3D1(txtUnit"+l.texture.slot+",lightDir,depth)";
	return ("1.0-"+str+"*"+l.biasname+".z;\n");
}

iv.shader.prototype.gen3dSampler1=function(ft,d){
	str="float Shadow3D1(samplerCube unit,vec3 dir,float depth){"+
	"return step(textureCube(unit,dir).r,depth);}"
	ft.push(str);
};

iv.shader.prototype.gen3dSampler=function(ft,d){
	
	var a=''+(d-1)/2,str=
	"float Shadow3D"+d+"(samplerCube unit,float scale,vec3 dir,vec3 point,vec3 lightOrg,vec3 normal,float depth,vec4 biasData){"+
	"vec3 pa,pb;if(abs(dir.z)<0.7)pa=cross(lightDir,vec3(0,0,1));else pa=cross(dir,vec3(0,1,0));"+
	"pb=cross(lightDir,pa);"+
	"float dot0=dot(point - lightOrg,normal);"+
	"float result=0.0,_depth,bias;"+
	"for(int x=-"+a+";x<="+a+";x++){"+
	"    vec3 a=pa*vec3(scale*float(x));"+
	"    for(int y=-"+a+";y<="+a+";y++){"+
	"    vec3 b=pb*vec3(scale*float(y));"+
	"    vec3 newDir=normalize(dir+b+a);"+
// we should calculate depth for each sample, instead of usign ready one	
	"    float denom=dot(newDir,normal);"+
	"    if(abs(denom)>0.001){"+
	"    float t = dot0 / denom;"+
	this.genDepth_("denom","biasData","t","_depth")+
		"}else _depth=depth;"+
	"    result+=step(textureCube(unit,newDir).r,_depth);"+
	"}}"+
	"return result/"+d*d+".0;}\n";
	ft.push(str);
}

iv.shader.prototype.fetchShadowI=function(l)
{
	var filter=l.light.light.shadowFilter,str;
	if(filter>1)
		str="Shadow"+filter+"(txtUnit"+l.texture.slot+","+l.vtmname+","+l.ptmname+",wPosition,normal,"+l.dirname+","+l.biasname+","+l.uvUname+","+l.uvVname+")";
	else
		str="Shadow1(txtUnit"+l.texture.slot+","+l.vtmname+","+l.ptmname+",wPosition,normal,"+l.dirname+","+l.biasname+")";
	return ("1.0-"+str+"*"+l.biasname+".z;");
};

// infinite light
iv.shader.prototype.gen2dSampler1=function(ft,spot)
{
	var depth=this.genDepth("lightDir","biasData",spot?"length(posInLight)":"posInLight.z","depth");
	var str="\nfloat Shadow1"+(spot?"spot":"")+"(sampler2D depths,mat4 vtm,mat4 ptm,vec4 wPosition,vec3 normal,vec3 lightDir,vec4 biasData)"+
	"{float depth;"+
	"vec3 posInLight=(vtm*vec4(wPosition.xyz,1.0)).xyz;"+
	"vec4 lightDevice= ptm* vec4(posInLight, 1.0);"+
	"vec2 lightUV=(lightDevice.xy/lightDevice.w)*0.5+0.5;";
	if(!spot)str+="if(lightUV.x>0.0 && lightUV.x<1.0 && lightUV.y>0.0 && lightUV.y<1.0){";
	str+=depth+	"return texture2DCompare(depths,lightUV,depth);}";
	if(!spot)str+="return 0.0;}";
	ft.push(str);
}

// spot light

iv.shader.prototype.fetchShadowS=function(l)
{
	var filter=l.light.light.shadowFilter,str;
	if(filter>1)
		str="Shadow"+filter+"spot(txtUnit"+l.texture.slot+","+l.vtmname+","+l.ptmname+",wPosition,normal,"+l.dirname+","+l.biasname+","+l.uvUname+","+l.uvVname+")";
	else
	var str="Shadow1spot(txtUnit"+l.texture.slot+","+l.vtmname+","+l.ptmname+",wPosition,normal,"+l.dirname+","+l.biasname+")";
	return ("1.0-"+str+"*"+l.biasname+".z;");
}

//specular sampler
iv.shader.prototype.genSSampler=function(ft,d)
{
	var depth=this.genDepth_("denom","biasData","length(posInLight)","depth");
	var a=''+(d-1)/2,str="\nfloat Shadow"+d+
	"spot(sampler2D depths,mat4 vtm,mat4 ptm,vec4 wPosition,vec3 normal,vec3 lightOrg,vec4 biasData,vec3 duU,vec3 duV)\n{float r=0.0,bias,depth;"+
	"float dot0=dot(wPosition.xyz - lightOrg,normal);"+
	"for(int x=-"+a+";x<="+a+";x++){\n"+
	"vec3 du=duU*vec3(x);\n"+
	"{for(int y=-"+a+"; y<="+a+"; y++){\n"+
	"vec3 duv=duV*vec3(y)+du;\n"+
	"vec3 pos=wPosition.xyz+duv;\n"+// new position in plane coplanar to texture plane
// calculate intersection of new dir with plane
	"vec3 newDir=normalize(pos-lightOrg);\n"+
	"float denom=dot(newDir,normal);\n"+
	"if(abs(denom)>0.0001){\n"+
	//"float t = dot(wPosition.xyz - lightOrg,normal) / denom;"+
	"float t = dot0 / denom;"+
	"pos = newDir*t+lightOrg;"+ // new position in plane coplanar with polygon plane

	"vec3 posInLight=(vtm*vec4(pos,1.0)).xyz;\n"+
	"vec4 lightDevice= ptm* vec4(posInLight, 1.0);\n"+
	"vec2 lightUV=(lightDevice.xy/lightDevice.w)*0.5+0.5;\n"+
	depth+
	"r+=texture2DCompare(depths,lightUV,depth);\n"+
	"}else {r+=1.0;}\n"+
	"}}}\n"+
	"return r/"+d*d+".0;}";
	ft.push(str);
}

iv.shader.prototype.genISampler=function(ft,d)
{	
	var depth=this.genDepth_("denom","biasData","(posInLight.z-dt)","depth");
	var a=''+(d-1)/2,str="\nfloat Shadow"+d+
	"(sampler2D depths,mat4 vtm,mat4 ptm,vec4 wPosition,vec3 normal,vec3 lightDir,vec4 biasData,vec3 duU,vec3 duV)\n{float r=0.0,bias,depth;"+
	"float denom=dot(lightDir,normal);\n"+
	"if(abs(denom)>0.0001){\n"+
	"for(int x=-"+a+";x<="+a+";x++){\n"+
	"vec3 du=duU*vec3(x);\n"+
	"{for(int y=-"+a+"; y<="+a+"; y++){\n"+
	"vec3 duv=duV*vec3(y)+du;\n"+
	"vec3 pos=wPosition.xyz+duv;\n"+
	"vec3 posInLight=(vtm*vec4(pos,1.0)).xyz;\n"+
	"vec4 lightDevice= ptm* vec4(posInLight, 1.0);\n"+
	"vec2 lightUV=(lightDevice.xy/lightDevice.w)*0.5+0.5;\n"+
	"if(lightUV.x>0.0 && lightUV.x<1.0 && lightUV.y>0.0 && lightUV.y<1.0){\n"+
	"float dt=(dot(lightDir-duv,normal)/denom-1.0);"+
	depth+
	"r+=texture2DCompare(depths,lightUV,depth);}else{r+=0.0;}\n"+
	"}}}}\n"+
	"return r/"+d*d+".0;}";
	ft.push(str);
}
/*
iv.shader.prototype.f1=function(ft,l) {
	ft.push("posInLight=("+l.vtmname+"* wPosition).xyz;");
	ft.push("lightDevice= "+l.ptmname+"* vec4(posInLight, 1.0);");
	ft.push("lightUV=(lightDevice.xy/lightDevice.w)*0.5+0.5;");
}*/

iv.shader.prototype.genDepth_=function(denom,bias,depth,rez)
{
	var str="bias=min("+denom+",0.0);"+rez+"=min("+depth+"*"+bias+".x+"+bias+".y-(1.0-bias)*"+bias+".w,0.999999);";
	return str;
}
iv.shader.prototype.genDepth=function(lDir,bias,depth,rez)
{
	//var str="bias=min(dot("+lDir+",normal),0.0);"+
	//rez+"=min("+depth+"*"+bias+".x+"+bias+".y-(1.0-bias)*0.02,0.999999);";
	return this.genDepth_("dot("+lDir+",normal)",bias,depth,rez);
}

iv.shader.prototype.f0=function(ft,l,str) {
	//ft.push("bias=min(dot(lightDir,normal),0.0);");
	//ft.push("depth=min("+str+"*"+l.biasname+".x+"+l.biasname+".y+bias*0.001,0.999999);");
	var _str=this.genDepth("lightDir",l.biasname,str,"depth");
	return ft.push(_str);
}

iv.shader.prototype.detach = function(gl)
{
	if(this.program!==null)
	{
		gl.detachShader(this.program,this.vShader);
		gl.detachShader(this.program,this.fShader);
		gl.deleteProgram(this.program);
		gl.deleteShader(this.vShader);
		gl.deleteShader(this.fShader);
	}
	this._reset();
}

// newObj means that we only updating shader
iv.shader.prototype.activate = function(space,info,flags,newObj)
{
	var mtl=this.mtl,gl=mtl.gl,i;
	if(!newObj){
	gl.useProgram(this.program);
	for(i=0;i<this.textures.length;i++)
		{
			var t=this.textures[i];
			if(t.txt.ready)
			{
				t.txt.bind(t.slot);
				var type=t.txt.target;
				if(type==gl.TEXTURE_2D && space.e_ans && (t.txt.filter==iv.FILTER_MIPMAP) )
					gl.texParameterf(type, space.e_ans.TEXTURE_MAX_ANISOTROPY_EXT, space.e_ansMax);
				t.txt.wrap(t.wrapS,t.wrapT);
				gl.uniform1i(t.uniform,t.slot);
			}
		}
	}
	var _i=this.vars;
	for(i=0;i<_i.length;i++)
	{
		var a=_i[i],s=a.slot;
		switch(a.ord)
		{
			case 4101:gl.uniformMatrix4fv(s, false, info.tm);break;
			case 4102: {
				var c=a.channel.color;
				if(flags&256) c=vec3.lerp(c,space.clrSelection,space.clrSelection[3]);
				gl.uniform3fv(s,c);
			} break;
			case 4103:gl.uniformMatrix3fv(s,false,a.channel.tm);break;
			case 4104:gl.uniform1f(s,a.channel.amount);break;
			case 4105:gl.uniform1f(s,-0.02);break;// positive - move backward
			case 4106:gl.uniform1f(s,info.opacity);break;

			default:
			if(!newObj)// update this only for fully new shader
			{
				switch(a.ord)
				{	
				case 4110:gl.uniform3fv(s,a.light.color);break;
				case 4111:gl.uniform3fv(s,a.light.org);break;
				case 4112:gl.uniform3fv(s,a.light.dir);break;
				case 4113:gl.uniform3fv(s,space.window.rcontext.view.from);break;
				case 4114:gl.uniformMatrix4fv(s,false,space.modelviewTM);break;
				case 4115:gl.uniformMatrix4fv(s,false,space.projectionTM);break;
				case 4116:gl.uniform1f(s,mtl.phong);break;
				case 4117:gl.uniform3fv(s,a.light.spot);break;

				case 4130:gl.uniformMatrix4fv(s,false,a.light.lightTm);break;
				case 4131:gl.uniformMatrix4fv(s,false,a.light.projectionTM);break;
				case 4132:gl.uniform4fv(s,a.light.shadowBias);break;
				case 4133:gl.uniform3fv(s,a.light.uvV);break;
				case 4134:gl.uniform3fv(s,a.light.uvU);break;


				}
			}
		}
	}
}

